/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use color_eyre::eyre::{Result, WrapErr};
use ratatui::{prelude::*, widgets::*};
use tokio::sync::mpsc::UnboundedSender;
use tokio::sync::OnceCell;

use super::Component;
use crate::{
  action::Action,
  config::{Config, KeyBindings},
};

pub struct Buflist {
  command_tx: Option<UnboundedSender<Action>>,
  config: Config,
}
impl Buflist {
  pub fn new(config: Config) -> Self {
    Buflist {
      command_tx: None,
      config,
    }
  }

  pub fn num_columns(&mut self, area: Rect, buffers: &crate::buffers::Buffers) -> u16 {
    let borders_height = 2;
    buffers
      .len()
      .div_ceil((area.height - borders_height).into())
      .try_into()
      .unwrap_or(u16::MAX)
  }

  pub fn borders_width(&self) -> u16 {
    2
  }
}

impl Component for Buflist {
  fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
    self.command_tx = Some(tx);
    Ok(())
  }

  fn update(&mut self, action: &Action) -> Result<Option<Action>> {
    Ok(None)
  }

  fn draw(
    &mut self,
    frame: &mut Frame<'_>,
    area: Rect,
    buffers: &crate::buffers::Buffers,
  ) -> Result<()> {
    let block = Block::new().borders(Borders::ALL);
    let inner_area = block.inner(area);
    block.render(area, frame.buffer_mut());

    let area = inner_area;

    let num_digits = u32::min(5, buffers.len().ilog10() + 1) as usize;
    let right_pad = " ".repeat(area.width.into());
    let mut stack = Vec::new(); // List of parent buffers of the current one
    let mut buffers_iter = buffers.iter().enumerate();
    let max_columns = u16::min(
      self.config.layout.buflist.max_columns,
      area.width.div_ceil(self.config.layout.buflist.column_width),
    );
    let column_width = self.config.layout.buflist.column_width;
    for col in 0..max_columns {
      let x = area.y + col * column_width;
      let allow_overflow = if col == max_columns - 2 {
        self.config.layout.buflist.penultimate_right_overflow
      } else if col == max_columns - 1 {
        // Always allow the last column to overflow
        true
      } else {
        false
      };
      for y in area.x..(area.x + area.height) {
        let Some((i, buf)) = buffers_iter.next() else {
          return Ok(());
        };
        match buffers
          .parents()
          .get(&buf.id())
          .into_iter()
          .flatten()
          .flat_map(|parent| stack.iter().position(|id| *id == *parent))
          .max() // if in multiple spaces, the last one is most likely to matter
        {
          Some(parent_position) => stack.truncate(parent_position + 1),
          None => stack.clear(),
        }
        stack.push(buf.id());
        let buf_number = format!("{}.", i + 1);
        let mut base_style = Style::default();
        if i == buffers.active_index() {
          base_style = base_style.on_blue();
        } else {
          base_style = base_style.bg(Color::Reset);
        }
        let tree_pad = " ".repeat(stack.len() - 1);
        let buf_number_style = base_style.green();
        let left_pad = " ".repeat((num_digits + 1).saturating_sub(buf_number.len()));
        frame.render_widget(
          Paragraph::new::<Line<'_>>(
            vec![
              Span::styled(left_pad, base_style),
              Span::styled(buf_number, buf_number_style),
              Span::styled(tree_pad, base_style),
              Span::styled(buf.short_name(), base_style),
              Span::styled(right_pad.clone(), base_style),
            ]
            .into(),
          ),
          Rect {
            x,
            y,
            width: if allow_overflow {
              area.width
            } else {
              column_width
            },
            height: 2,
          }
          .intersection(area),
        );
      }
    }

    // TODO: Do something with the remaining buffers, if any. eg. print "XX more buffers"
    // at the end or display a scrollbar

    Ok(())
  }
}
